@@ -1,4 +1,4 @@
-/*	$OpenBSD: printconf.c,v 1.159 2022/09/21 21:12:04 claudio Exp $	*/
+/*	$OpenBSD: printconf.c,v 1.167 2023/04/28 13:23:52 claudio Exp $	*/
 
 /*
  * Copyright (c) 2003, 2004 Henning Brauer <henning@openbsd.org>
@@ -38,10 +38,12 @@ void		 print_l3vpn_targets(struct filter_set_head *, const char *);
 void		 print_l3vpn(struct l3vpn *);
 const char	*print_af(uint8_t);
 void		 print_network(struct network_config *, const char *);
+void		 print_flowspec(struct flowspec_config *, const char *);
 void		 print_as_sets(struct as_set_head *);
 void		 print_prefixsets(struct prefixset_head *);
 void		 print_originsets(struct prefixset_head *);
 void		 print_roa(struct roa_tree *);
+void		 print_aspa(struct aspa_tree *);
 void		 print_rtrs(struct rtr_config_head *);
 void		 print_peer(struct peer_config *, struct bgpd_config *,
 		    const char *);
@@ -217,6 +219,8 @@ print_community(struct community *c)
 		switch (type) {
 		case EXT_COMMUNITY_TRANS_TWO_AS:
 		case EXT_COMMUNITY_TRANS_FOUR_AS:
+		case EXT_COMMUNITY_GEN_TWO_AS:
+		case EXT_COMMUNITY_GEN_FOUR_AS:
 			if ((c->flags >> 8 & 0xff) == COMMUNITY_NEIGHBOR_AS)
 				printf("neighbor-as:");
 			else if ((c->flags >> 8 & 0xff) == COMMUNITY_LOCAL_AS)
@@ -225,6 +229,7 @@ print_community(struct community *c)
 				printf("%s:", log_as(c->data1));
 			break;
 		case EXT_COMMUNITY_TRANS_IPV4:
+		case EXT_COMMUNITY_GEN_IPV4:
 			addr.s_addr = htonl(c->data1);
 			printf("%s:", inet_ntoa(addr));
 			break;
@@ -234,6 +239,9 @@ print_community(struct community *c)
 		case EXT_COMMUNITY_TRANS_TWO_AS:
 		case EXT_COMMUNITY_TRANS_FOUR_AS:
 		case EXT_COMMUNITY_TRANS_IPV4:
+		case EXT_COMMUNITY_GEN_TWO_AS:
+		case EXT_COMMUNITY_GEN_FOUR_AS:
+		case EXT_COMMUNITY_GEN_IPV4:
 			if ((c->flags >> 16 & 0xff) == COMMUNITY_ANY)
 				printf("* ");
 			else if ((c->flags >> 16 & 0xff) ==
@@ -460,12 +468,12 @@ print_af(uint8_t aid)
 {
 	/*
 	 * Hack around the fact that aid2str() will return "IPv4 unicast"
-	 * for AID_INET. AID_INET and AID_INET6 need special handling and
-	 * the other AID should never end up here (at least for now).
+	 * for AID_INET. AID_INET, AID_INET6 and the flowspec AID need
+	 * special handling and the other AID should never end up here.
 	 */
-	if (aid == AID_INET)
+	if (aid == AID_INET || aid == AID_FLOWSPECv4)
 		return ("inet");
-	if (aid == AID_INET6)
+	if (aid == AID_INET6 || aid == AID_FLOWSPECv6)
 		return ("inet6");
 	return (aid2str(aid));
 }
@@ -502,6 +510,115 @@ print_network(struct network_config *n, const char *c)
 	printf("\n");
 }
 
+static void
+print_flowspec_list(struct flowspec *f, int type, int is_v6)
+{
+	const uint8_t *comp;
+	const char *fmt;
+	int complen, off = 0;
+
+	if (flowspec_get_component(f->data, f->len, type, is_v6,
+	    &comp, &complen) != 1)
+		return;
+
+	printf("%s ", flowspec_fmt_label(type));
+	fmt = flowspec_fmt_num_op(comp, complen, &off);
+	if (off == -1) {
+		printf("%s ", fmt);
+	} else {
+		printf("{ %s ", fmt);
+		do {
+			fmt = flowspec_fmt_num_op(comp, complen, &off);
+			printf("%s ", fmt);
+		} while (off != -1);
+		printf("} ");
+	}
+}
+
+static void
+print_flowspec_flags(struct flowspec *f, int type, int is_v6)
+{
+	const uint8_t *comp;
+	const char *fmt, *flags;
+	int complen, off = 0;
+
+	switch (type) {
+	case FLOWSPEC_TYPE_TCP_FLAGS:
+		flags = FLOWSPEC_TCP_FLAG_STRING;
+		break;
+	case FLOWSPEC_TYPE_FRAG:
+		if (!is_v6)
+			flags = FLOWSPEC_FRAG_STRING4;
+		else
+			flags = FLOWSPEC_FRAG_STRING6;
+		break;
+	default:
+		printf("??? ");
+		return;
+	}
+
+	if (flowspec_get_component(f->data, f->len, type, is_v6,
+	    &comp, &complen) != 1)
+		return;
+
+	printf("%s ", flowspec_fmt_label(type));
+
+	fmt = flowspec_fmt_bin_op(comp, complen, &off, flags);
+	if (off == -1) {
+		printf("%s ", fmt);
+	} else {
+		printf("{ %s ", fmt);
+		do {
+			fmt = flowspec_fmt_bin_op(comp, complen, &off, flags);
+			printf("%s ", fmt);
+		} while (off != -1);
+		printf("} ");
+	}
+}
+
+static void
+print_flowspec_addr(struct flowspec *f, int type, int is_v6)
+{
+	struct bgpd_addr addr;
+	uint8_t plen;
+
+	flowspec_get_addr(f->data, f->len, type, is_v6, &addr, &plen, NULL);
+	if (plen == 0)
+		printf("%s any ", flowspec_fmt_label(type));
+	else
+		printf("%s %s/%u ", flowspec_fmt_label(type),
+		    log_addr(&addr), plen);
+}
+
+void
+print_flowspec(struct flowspec_config *fconf, const char *c)
+{
+	struct flowspec *f = fconf->flow;
+	int is_v6 = (f->aid == AID_FLOWSPECv6);
+
+	printf("%sflowspec %s ", c, print_af(f->aid));
+
+	print_flowspec_list(f, FLOWSPEC_TYPE_PROTO, is_v6);
+
+	print_flowspec_addr(f, FLOWSPEC_TYPE_SOURCE, is_v6);
+	print_flowspec_list(f, FLOWSPEC_TYPE_SRC_PORT, is_v6);
+
+	print_flowspec_addr(f, FLOWSPEC_TYPE_DEST, is_v6);
+	print_flowspec_list(f, FLOWSPEC_TYPE_DST_PORT, is_v6);
+
+	print_flowspec_list(f, FLOWSPEC_TYPE_DSCP, is_v6);
+	print_flowspec_list(f, FLOWSPEC_TYPE_PKT_LEN, is_v6);
+	print_flowspec_flags(f, FLOWSPEC_TYPE_TCP_FLAGS, is_v6);
+	print_flowspec_flags(f, FLOWSPEC_TYPE_FRAG, is_v6);
+
+	/* TODO: fixup the code handling to be like in the parser */
+	print_flowspec_list(f, FLOWSPEC_TYPE_ICMP_TYPE, is_v6);
+	print_flowspec_list(f, FLOWSPEC_TYPE_ICMP_CODE, is_v6);
+
+	print_set(&fconf->attrset);
+	printf("\n");
+}
+
 void
 print_as_sets(struct as_set_head *as_sets)
 {
@@ -590,6 +707,32 @@ print_roa(struct roa_tree *r)
 	printf("\n}\n\n");
 }
 
+void
+print_aspa(struct aspa_tree *a)
+{
+	struct aspa_set	*aspa;
+	uint32_t i;
+
+	if (RB_EMPTY(a))
+		return;
+
+	printf("aspa-set {");
+	RB_FOREACH(aspa, aspa_tree, a) {
+		printf("\n\t");
+		printf("customer-as %s", log_as(aspa->as));
+		if (aspa->expires != 0)
+			printf(" expires %lld", (long long)aspa->expires);
+		printf(" provider-as { ");
+		for (i = 0; i < aspa->num; i++) {
+			printf("%s ", log_as(aspa->tas[i]));
+			if (aspa->tas_aid[i] != AID_UNSPEC)
+				printf("%s ", print_af(aspa->tas_aid[i]));
+		}
+		printf("}");
+	}
+	printf("\n}\n\n");
+}
+
 void
 print_rtrs(struct rtr_config_head *rh)
 {
@@ -643,6 +786,8 @@ print_peer(struct peer_config *p, struct bgpd_config *conf, const char *c)
 		    log_addr(&p->local_addr_v6));
 	if (p->remote_port != BGP_PORT)
 		printf("%s\tport %hu\n", c, p->remote_port);
+	if (p->role != ROLE_NONE)
+		printf("%s\trole %s\n", c, log_policy(p->role));
 	if (p->max_prefix) {
 		printf("%s\tmax-prefix %u", c, p->max_prefix);
 		if (p->max_prefix_restart)
@@ -818,12 +963,10 @@ print_announce(struct peer_config *p, const char *c)
 			printf(" max %d", p->eval.maxpaths);
 		printf("\n");
 	}
-	if (p->capabilities.role_ena) {
-		printf("%s\tannounce policy %s%s\n", c,
-		    log_policy(p->capabilities.role),
-		    p->capabilities.role_ena == 2 ? " enforce" : "");
+	if (p->capabilities.policy) {
+		printf("%s\tannounce policy %s\n", c,
+		    p->capabilities.policy == 2 ? "enforce" : "yes");
 	}
-
 }
 
 void
@@ -917,6 +1060,22 @@ print_rule(struct bgpd_config *conf, struct filter_rule *r)
 		}
 	}
 
+	if (r->match.avs.is_set) {
+		switch (r->match.avs.validity) {
+		case ASPA_VALID:
+			printf("avs valid ");
+			break;
+		case ASPA_INVALID:
+			printf("avs invalid ");
+			break;
+		case ASPA_UNKNOWN:
+			printf("avs unknown ");
+			break;
+		default:
+			printf("avs ??? %d ??? ", r->match.avs.validity);
+		}
+	}
+
 	if (r->match.prefix.addr.aid != AID_UNSPEC) {
 		printf("prefix ");
 		print_prefix(&r->match.prefix);
@@ -1090,17 +1249,21 @@ print_config(struct bgpd_config *conf, struct rib_names *rib_l)
 {
 	struct filter_rule	*r;
 	struct network		*n;
+	struct flowspec_config	*f;
 	struct rde_rib		*rr;
 	struct l3vpn		*vpn;
 
 	print_mainconf(conf);
 	print_rtrs(&conf->rtrs);
 	print_roa(&conf->roa);
+	print_aspa(&conf->aspa);
 	print_as_sets(&conf->as_sets);
 	print_prefixsets(&conf->prefixsets);
 	print_originsets(&conf->originsets);
 	TAILQ_FOREACH(n, &conf->networks, entry)
 		print_network(&n->net, "");
+	RB_FOREACH(f, flowspec_tree, &conf->flowspecs)
+		print_flowspec(f, "");
 	if (!SIMPLEQ_EMPTY(&conf->l3vpns))
 		printf("\n");
 	SIMPLEQ_FOREACH(vpn, &conf->l3vpns, entry)
